--------------------------------------------------------------------------------
-- Locale
--

local L = setmetatable({}, {__index = function(self, key) self[key] = key return key end })

--------------------------------------------------------------------------------
-- DKP-type specific variables
--

local FLASK_MULTIPLIER = 0.0
local ELIXIR_MULTIPLIER = 0.0
local FOOD_MULTIPLIER = 0.0

local flaskIds = {
	92679, 94160, 79469, 79470, 79471, 79472,
}
local elixirIds = {
	79480, 79631, 89344, 89345,
	79481, 79632, 79477, 79635, 79474, 79468, 89343, 89347, 89342, 89346,
	-- WotLK
	--60340, 80532, 60341, 60344, 60346, 28497, 53748, 53749, 33721, 53746,
	--53751, 53764, 60347, 53763, 53747, 60343,
}
local flasks = setmetatable({}, {__index = function(self, key)
	for i, id in next, flaskIds do
		local n = GetSpellInfo(id)
		if n then self[n] = true end
	end
	flaskIds = nil
	setmetatable(self, nil)
	return self[key]
end})
local elixirs = setmetatable({}, {__index = function(self, key)
	for i, id in next, elixirIds do
		local n = GetSpellInfo(id)
		if n then self[n] = true end
	end
	elixirIds = nil
	setmetatable(self, nil)
	return self[key]
end})

local foods = {
	[(GetSpellInfo(44100))] = true,
	[(GetSpellInfo(19706))] = true,
}

local ignoredItems = {
	[30316] = true, -- Devastation
	[30317] = true, -- Cosmic Infuser
	[30312] = true, -- Infinity Blade
	[30318] = true, -- Netherstrand Longbow
	[30313] = true, -- Staff of Disintegration
	[30311] = true, -- Warp Slicer
	[30320] = true, -- Bundle of Nether Spikes
	[30314] = true, -- Phaseshift Bulwark
	[30319] = true, -- Nether Spikes
	[29434] = true, -- Badge of Justice
	[22450] = true, -- Void Crystal
	[20725] = true, -- Nexus Crystal

	[32897] = true, -- Mark of the Illidari
	[32428] = true, -- Heart of Darkness
	[34664] = true, -- Sunmote

	[32230] = true, -- Shadowsong Amethyst
	[32249] = true, -- Seaspray Emerald
	[32231] = true, -- Pyrestone
	[32229] = true, -- Lionseye
	[32228] = true, -- Empyrean Sapphire
	[32227] = true, -- Crimson Spinel

	[36919] = true, -- Cardinal Ruby
	[36925] = true, -- Majestic Zircon
	[36922] = true, -- King's Amber
	[36928] = true, -- Dreadstone
	[36934] = true, -- Eye of Zul
	[36931] = true, -- Ametrine

	[34057] = true, -- Abyss Crystal
	[50274] = true, -- Shadowfrost Shard

	[52722] = true, -- Maelstrom Crystal
	[69815] = true, -- Seething Cinder
    [77952] = true, -- Elementium Gem Cluster
    [77951] = true, -- Shadowy Gem
    [78352] = true, -- Fragment of Deathwing's Jaw
    [78890] = true, -- Crystalline Geode
    [78891] = true, -- Elementium-Coated Geode
    [71998] = true, -- Essence of Destruction
}

local icon = LibStub("LibDBIcon-1.0", true)
local JSON = LibStub("LibJSON-1.0", true)

local db = nil

local halfHour = 60 * 30
local oneHour = halfHour * 2

local handle_HourTicker = nil
local handle_MinuteTicker = nil
local handle_Broadcast = nil
local handle_BidTimer = nil

local cache_AttendancePercent = {}

local guildRanks = {}
local guildMemberRanks = {}
local guildMemberList = {}

local lootDialogOpen = nil
local lootQueue = {}
local updateDisplayedNames = true
local takingBids = nil
local tmpBidTable = {}
local bidders = {}
local playerName = UnitName("player")

local hexColors = {}
local wtfColor = "|cffa0a0a0%s|r"
for k, v in pairs(RAID_CLASS_COLORS) do
	hexColors[k] = "|cff" .. string.format("%02x%02x%02x", v.r * 255, v.g * 255, v.b * 255) .. "%s|r"
end
local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = guildMemberList[key] or select(2, UnitClass(key)) or "WTF"
		if not class or class == "WTF" then
			return wtfColor:format(key)
		else
			self[key] = hexColors[class]:format(key)
			return self[key]
		end
	end
})

--------------------------------------------------------------------------------
-- Utility functions
--

local function debug(t)
	if db.DEBUG then print("[DEBUG] " .. t) end
end

local function broadcastMulti()
	debug("broadcastMulti()")
	local dkp = PlantainDKP:Serialize(db.dkp)
	local att = {}
	for k, v in pairs(db.attendance) do
		if not v:find("1") then
			att[k] = "-"
		elseif not v:find("0") then
			att[k] = "!"
		else
			att[k] = v
		end
	end
	local attendance = PlantainDKP:Serialize(att)
	-- Broadcast attendance first, since it takes aaaages to send.
	PlantainDKP:SendCommMessage("BDKP-ATT-M", attendance, "GUILD")
	PlantainDKP:SendCommMessage("BDKP-DKP-M", dkp, "GUILD")
	wipe(att)
end

local function broadcastSingle(name)
	debug("broadcastSingle(" .. name .. ")")
	if db.attendance[name] then
		local attendance = PlantainDKP:Serialize({ name, db.attendance[name] })
		PlantainDKP:SendCommMessage("BDKP-ATT-S", attendance, "GUILD")
	end
	if db.dkp[name] then
		local dkp = PlantainDKP:Serialize({ name, db.dkp[name] })
		PlantainDKP:SendCommMessage("BDKP-DKP-S", dkp, "GUILD")
	end
end

-- Thanks, Wobin.
local function round(n, shift)
	shift = 10 ^ shift
	return math.floor((n * shift) + 0.5) / shift
end

local function getDKPMultiplier(player)
	local multiplier = 1
	local buffIndex = 1
	local buff = UnitBuff(player, buffIndex)
	while buff do
		if flasks[buff] then
			multiplier = multiplier + FLASK_MULTIPLIER
		elseif foods[buff] then
			multiplier = multiplier + FOOD_MULTIPLIER
		elseif elixirs[buff] then
			multiplier = multiplier + ELIXIR_MULTIPLIER
		end
		buffIndex = buffIndex + 1
		buff = UnitBuff(player, buffIndex)
	end
	return multiplier
end

local function getAttendancePercent(name)
	if cache_AttendancePercent[name] then return cache_AttendancePercent[name] end
	if not db.attendance[name] then return 0 end
	local total = 0
	for k in db.attendance[name]:gmatch("%d") do
		if k == "1" then total = total + 1 end
	end
	local p = (total / db.attendanceAverageOver) * 100
	cache_AttendancePercent[name] = p
	return p
end

local function getDKP(name)
	if not name then return end
	if not db.dkp[name] then return end
	return round(db.dkp[name], 2)
end

local function getDKPPercent(name)
	local p = getAttendancePercent(name)
	if p >= 60 then return 100
	else return (1.5 * p) + 10 end
end

local function getModifiedDKP(name)
	if not name then return end
	if not db.dkp[name] then return end
	local p = getDKPPercent(name)
	if p == 100 then return round(db.dkp[name], 2)
	else return round((db.dkp[name] * p) / 100, 2)
	end
end

local function attendanceSorter(a, b)
	local aA = getAttendancePercent(a) or 0
	local bA = getAttendancePercent(b) or 0
	return aA > bA
end

local function dkpSorter(a, b)
	local aDKP = getModifiedDKP(a) or 0
	local bDKP = getModifiedDKP(b) or 0
	return aDKP > bDKP
end

local function itemLooted(player, itemLink)
	if db.mode == 2 then return end
	if lootDialogOpen then
		table.insert(lootQueue, {player, itemLink})
	else
		lootDialogOpen = true
		StaticPopupDialogs["PlantainDKPItemQuery"].text = L["%s looted %s, click Add to add it to his loot table."]:format(coloredName[player], itemLink)
		local popup = StaticPopup_Show("PlantainDKPItemQuery")
		if popup then
			popup.data = {player, itemLink}
		end
	end
end
local function timerLoot(tbl) itemLooted(unpack(tbl)) end
local function popLootQueue()
	if #lootQueue > 0 then
		local n = table.remove(lootQueue, 1)
		PlantainDKP:ScheduleTimer(timerLoot, 1, n)
	end
end
local function ignoreItem()
	lootDialogOpen = nil
	popLootQueue()
end

--------------------------------------------------------------------------------
-- Options
--

PlantainDKP = LibStub("AceAddon-3.0"):NewAddon("PlantainDKP", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0", "AceSerializer-3.0")
local PlantainDKP = PlantainDKP
local BananaDKP = PlantainDKP

local function notHost() return db.mode ~= 1 end

local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("PlantainDKP", {
	type = "data source",
	text = "PDKP",
	icon = "Interface\\Icons\\inv_misc_food_24",
	OnClick = function(self, button)
		if button == "LeftButton" then
			if db.tooltipSortMethod == "dkp" then
				db.tooltipSortMethod = "attendance"
			else
				db.tooltipSortMethod = "dkp"
			end
			updateDisplayedNames = true
		else
			InterfaceOptionsFrame_OpenToCategory("PlantainDKP")
		end
	end
})

--------------------------------------------------------------------------------
-- Slash command
--

local slashHandlers = {}
if icon then
	function slashHandlers.minimap()
		if db.minimap.hide then
			icon:Show("PlantainDKP")
			db.minimap.hide = nil
            print(L["|cff2476ffPDKP:|r Minimap icon is SHOWN."])
		else
			icon:Hide("PlantainDKP")
			db.minimap.hide = true
            print(L["|cff2476ffPDKP:|r Minimap icon is HIDDEN."])
		end
	end
end
function slashHandlers.loot(input)
	if notHost() then return end
	if not db.dkp[input] then
		print(L["|cff2476ffPDKP:|r %s |cffff4488doesn't exist|r in the DKP database."]:format(input))
		return
	end
	PlantainDKP:AddItem(input)
end
function slashHandlers.unloot(input)
	if notHost() then return end
	if not db.dkp[input] then
		print(L["|cff2476ffPDKP:|r %s |cffff4488doesn't exist|r in the DKP database."]:format(input))
		return
	end
	local dkp = db.dkp[input]
    local adjust = 0
    adjust = dkp / 75 * 25
    local newDkp = dkp + adjust
    db.dkp[input] = newDkp
    broadcastSingle(input)
	print(L["|cff2476ffPDKP:|r %s' DKP has been adjusted by |cffff4488%d|r."]:format(input, adjust))
end
function slashHandlers.sitout(input)
    if notHost() then return end
	if not db.dkp[input] then
		print(L["|cff2476ffPDKP:|r %s |cffff4488doesn't exist|r in the DKP database."]:format(input))
		return
	end
    PlantainDKP:ProcessSitoutWhisper(input)
end
function slashHandlers.debug()
    if db.DEBUG == true then
        db.DEBUG = false
        print(L["|cff2476ffPDKP:|r DEBUG: OFF"])
    else
        db.DEBUG = true
        print(L["|cff2476ffPDKP:|r DEBUG: ON"])
    end
end
function slashHandlers.chatloot()
    if notHost() then return end
    if db.chatloot == true then
        db.chatloot = false
        print(L["|cff2476ffPDKP:|r Will NOT broadcast loot in Guild Chat!"])
    else
        db.chatloot = true
        print(L["|cff2476ffPDKP:|r Will broadcast loot in Guild Chat!"])
    end
end
function slashHandlers.wipe(input)
    if notHost() then return end
	if not db.dkp[input] then
		print(L["|cff2476ffPDKP:|r %s |cffff4488doesn't exist|r in the DKP database."]:format(input))
		return
	end
    db.dkp[input] = nil
    print(L["|cff2476ffPDKP:|r DKP for %s was wiped."]:format(input))
    return
end
function slashHandlers.inc(input)
	if notHost() then return end
	PlantainDKP:IncrementHourCounter(input)
end
function slashHandlers.dec(input)
	if notHost() then return end
	PlantainDKP:DecrementHourCounter(input)
end
function slashHandlers.role(input)
    if handle_HourTicker or handle_MinuteTicker then
        print(L["|cff2476ffPDKP:|r You can't change your own role while the DKP system is active."])
        return
    end
    PlantainDKP:CancelTimer(handle_Broadcast, true)
    handle_Broadcast = nil
    
	if input == "server" then
		db.mode = 1
		print(L["|cff2476ffPDKP:|r You are now running PlantainDKP in |cffff4488server|r mode."])
	else
		if not input or tostring(input):trim():len() == 0 then
			print(L["|cff2476ffPDKP:|r 'role' accepts either 'server' or the name of the player you want to use as the host."])
            if db.mode == 2 then
                print(L["|cff2476ffPDKP:|r You are running PlantainDKP in |cff44ff44client|r mode, using %s as the server."]:format(coloredName[db.host]))
            else
                print(L["|cff2476ffPDKP:|r You are running PlantainDKP in |cff44ff44server|r mode"])
            end
			return
		end
        
        if guildMemberList[input] then
            db.mode = 2
            db.host = input
            print(L["|cff2476ffPDKP:|r You are now running PlantainDKP in |cff44ff44client|r mode, using %s as the server."]:format(input))
        else
            print(L["|cff2476ffPDKP:|r There is no player named %s in the guild."]:format(input)) 
        end
	end
end
function slashHandlers.broadcast(input)
    if notHost() then return end
    if input == "yes" then
        print(L["|cff2476ffPDKP:|r Broadcasting DKP & Attendance now."])
        broadcastMulti()
    else
        print(L["|cff2476ffPDKP:|r Are you sure you want to broadcast DKP & Attendance now?"])
    end
end
function slashHandlers.req(input)
    if not notHost() then return end
    if input == "yes" then
        if db.host then
            print(L["|cff2476ffPDKP:|r Requesting DKP & Attendance from %s now."]:format(db.host))
            PlantainDKP:SendCommMessage("BDKP-REQ", db.host, "GUILD")
        else
            print(L["|cff2476ffPDKP:|r We don't have a server to sync with. Set it with /pdkp role <leader>"])
        end
    else
        print(L["|cff2476ffPDKP:|r Are you sure you want to request DKP & Attendance now?"])
    end
end
function slashHandlers.start()
	if notHost() or not UnitInRaid("player") or handle_HourTicker then return end
	PlantainDKP:StartCounter()
end
function slashHandlers.stop()
	if notHost() or not handle_HourTicker then return end
	PlantainDKP:StopCounter()
end

function slashHandlers.export()
    if notHost() then return end
        local object = { nick = nil }
        
        for name, attendance in pairs(db.attendance) do
            local dkp, hour = db.dkp[name], db.hours[name]
            object[name] = {dkp, hour, attendance}
        end
        
        local dkpexport = JSON.Serialize(object)
        
        StaticPopupDialogs["EXPORT"] = {
        text = "DKP Export String",
        button1 = "Thank you!",
        timeout = 0,
        whileDead = true,
        hideOnEscape = true,
        OnShow = function (self, data)
        self.editBox:SetText(dkpexport)
        end,
        hasEditBox = true,
        }
        StaticPopup_Show("EXPORT")
end

local function handleSlashCmd(input)
	local arg = {strsplit(" ", input)}
	local command = table.remove(arg, 1)
	if command and slashHandlers[command] then
		slashHandlers[command](unpack(arg))
	else
		local tbl = {}
		for k in pairs(slashHandlers) do tbl[#tbl + 1] = k end
		print(L["|cff2476ffPDKP:|r Unknown slash command %s. Accepted commands are %s."]:format(tostring(command), table.concat(tbl, ", ")))
	end
end

--------------------------------------------------------------------------------
-- Initialization
--

function PlantainDKP:OnInitialize()
	if not _G.BananaDKPDB then _G.BananaDKPDB = {} end
	db = _G.BananaDKPDB
	for k, v in pairs({
		sitouts = {}, -- Remember this across sessions.
		mode = 2, -- 1 = server, 2 = client
		host = nil, -- No host set by default.
		disallowRanks = {},
		disallowClasses = {},
		dkp = {},
		hours = {},
		totalRaidHours = 0,
		attendance = {},
		attendanceAverageOver = 45,
		itemsLooted = {},
		minimap = {},
		lastDKPUpdate = nil,
		lastReload = 0,
		penalties = {},
		tooltipSortMethod = "dkp",
        DEBUG = false,
        chatloot = true,
	}) do
		if not db[k] then db[k] = v end
	end
    
	-- In case the attendanceAverageOver value has changed,
	-- update attendance accordingly.
	local aAO = db.attendanceAverageOver
	for k, v in pairs(db.attendance) do
		-- Attendance tracked for longer than aAO, so trim it
		if #v > aAO then
			-- Shave from the old, since the recent data is probably more
			-- accurate
			db.attendance[k] = v:sub(1, aAO)

		-- Attendance not tracked for as long as aAO, so append
		elseif #v < aAO then
			local d = aAO - #v
			-- Add the new hours at the end, in case it's a slacker,
			-- he'll drop faster back down on attendance.
			db.attendance[k] = v .. ("1"):rep(d)
		end
	end

	_G.SlashCmdList.PlantainDKP = handleSlashCmd
	_G.SLASH_PlantainDKP1 = "/bananadkp"
	_G.SLASH_PlantainDKP2 = "/bdkp"
    _G.SLASH_PlantainDKP3 = "/pdkp"
    _G.SLASH_PlantainDKP4 = "/plantaindkp"

	if icon then
		icon:Register("PlantainDKP", LDB, db.minimap)
	end

	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", function(self, event, msg)
		if msg == "sitout" or msg:find("^bdkp") or msg:find("^pdkp") or (takingBids and msg:lower() == "bid") then
			return true
		end
	end)

	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", function(self, event, msg)
		if msg:find("^BDKP") or msg:find("^PDKP") then return true end
	end)

	if IsInGuild() then GuildRoster() end
end

function PlantainDKP:OnEnable()
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end
	if type(popup["PlantainDKPItemQuery"]) ~= "table" then
		popup["PlantainDKPItemQuery"] = {
			button1 = L["Add"],
			button2 = L["Ignore"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 0,
			OnCancel = ignoreItem,
			OnAccept = function(self, data)
				PlantainDKP:AddItem(unpack(data))
			end,
			preferredIndex = STATICPOPUP_NUMDIALOGS,
		}
	end

	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")

	self:RegisterEvent("CHAT_MSG_RAID")
	self:RegisterEvent("CHAT_MSG_RAID_LEADER")
	self:RegisterEvent("CHAT_MSG_WHISPER")

	self:RegisterComm("BDKP-REQ",   "OnRequestComm")
	self:RegisterComm("BDKP-DKP-M", "OnMultiDKPComm")
	self:RegisterComm("BDKP-ATT-M", "OnMultiAttendanceComm")
	self:RegisterComm("BDKP-DKP-S", "OnSingleDKPComm")
	self:RegisterComm("BDKP-ATT-S", "OnSingleAttendanceComm")

	if not IsInGuild() then return end

	-- Throttle this in case we're reloading often, which we do when coding addons!
	local t = GetTime()
	if t < (db.lastReload + 600) then return end
	db.lastReload = t

	if db.mode == 1 then
		broadcastMulti()
		handle_Broadcast = self:ScheduleRepeatingTimer(broadcastMulti, oneHour)
	elseif db.host and (not db.lastDKPUpdate or (time() > db.lastDKPUpdate + oneHour)) then
		self:SendCommMessage("BDKP-REQ", db.host, "GUILD")
	end
end

--------------------------------------------------------------------------------
-- DKP Penalty handling
--

local penaltyItems = {
	[61031] = L["Toy Train Set"],
	[49844] = L["Direbrew's Remote"],
}

local penaltySpells = {
    [61716] = L["Blossoming Branch"],
    [24740] = L["Hallowed Wand - Wisp"],
    [24732] = L["Hallowed Wand - Bat"],
}

function PlantainDKP:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, _, _, source, _, _, _, _, _, _, spellId)
	if not handle_HourTicker then return end -- We're not running
    
	if event == "SPELL_CREATE" and penaltyItems[spellId] and UnitInRaid(source) then

		if not db.penalties[source] then
			db.penalties[source] = 1
		else
			db.penalties[source] = db.penalties[source] + 1
		end

		db.dkp[source] = nil
		if db.penalties[source] < 5 then
			self:Chat(L["PDKP: %s used a %s and has received a 100%% DKP penalty."]:format(source, penaltyItems[spellId]), "GUILD")
		else
			self:Chat(L["PDKP: %s has gotten %d penalties so far. Perhaps it's time he is removed from the guild."]:format(source), "GUILD")
		end
    elseif event == "SPELL_AURA_APPLIED" and penaltySpells[spellId] and UnitInRaid(source) then
		if not db.penalties[source] then
			db.penalties[source] = 1
		else
			db.penalties[source] = db.penalties[source] + 1
		end

		db.dkp[source] = nil
		if db.penalties[source] < 5 then
			self:Chat(L["PDKP: %s used a %s and has received a 100%% DKP penalty."]:format(source, penaltySpells[spellId]), "GUILD")
		else
			self:Chat(L["PDKP: %s has gotten %d penalties so far. Perhaps it's time he is removed from the guild."]:format(source), "GUILD")
		end
    end

end

--------------------------------------------------------------------------------
-- Comm handling
--

function PlantainDKP:OnRequestComm(prefix, message, distribution, sender)
	if notHost() or message ~= playerName then return end
    debug(sender .. " is requesting DKP info. Sending...")
	debug("OnRequestComm(" .. sender .. ")")
	broadcastMulti()
end

function PlantainDKP:OnMultiDKPComm(prefix, message, distribution, sender)
	debug("OnMultiDKPComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, dkp = self:Deserialize(message)
	if not ret then error(dkp) end
	for k, v in pairs(dkp) do
		db.dkp[k] = v
	end
	wipe(dkp)
	db.lastDKPUpdate = time()
end

function PlantainDKP:OnSingleDKPComm(prefix, message, distribution, sender)
	debug("OnSingleDKPComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, tbl = self:Deserialize(message)
	if not ret then error(tbl) end
	db.dkp[tbl[1]] = tbl[2]
	db.lastDKPUpdate = time()
end

function PlantainDKP:OnMultiAttendanceComm(prefix, message, distribution, sender)
	debug("OnMultiAttendanceComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, attendance = self:Deserialize(message)
	if not ret then error(attendance) end
	for k, v in pairs(attendance) do
		if v == "-" then -- All 0's
			db.attendance[k] = ("0"):rep(db.attendanceAverageOver)
		elseif v == "!" then -- All 1's
			db.attendance[k] = ("1"):rep(db.attendanceAverageOver)
		else
			db.attendance[k] = v
		end
		cache_AttendancePercent[k] = nil
	end
	wipe(attendance)
	db.lastDKPUpdate = time()
end

function PlantainDKP:OnSingleAttendanceComm(prefix, message, distribution, sender)
	debug("OnSingleAttendanceComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, tbl = self:Deserialize(message)
	if not ret then error(tbl) end
	db.attendance[tbl[1]] = tbl[2]
	cache_AttendancePercent[tbl[1]] = nil
	db.lastDKPUpdate = time()
end

--------------------------------------------------------------------------------
-- Ticker handling
--

local function storeTickerCounter(timeLeft)
	db.storedCounter = timeLeft - 60
	handle_MinuteTicker = PlantainDKP:ScheduleTimer(storeTickerCounter, 60, db.storedCounter)
end

function PlantainDKP:StartCounter()
	if handle_HourTicker then
		print("We wanted to start the counter, but it's already running!")
		return
	end

    if notHost() then
        print("We tried to start the hour ticker, but we're not configured as a host.")
        return
    end

	LDB.text = "|cffff0000[!]|rPDKP"

	local nextUpdate = db.storedCounter or (math.random(10, 30) * 60)
	handle_HourTicker = self:ScheduleTimer("IncrementHourCounter", nextUpdate)

	if db.storedCounter then
		print(L["|cff2476ffPDKP:|r Starting hour counter from last session. Time until next increment is |cff44ff44%d minutes|r."]:format(nextUpdate / 60))
	else
		print(L["|cff2476ffPDKP:|r Automatic hour counting has started, the first increment will be in |cff44ff44%d minutes|r."]:format(nextUpdate / 60))
		self:Chat(L["DKP hour counter started, whisper me 'sitout' if you are not in the raid but available on standby."], "GUILD")
		for player in pairs(db.sitouts) do
			self:Chat(L["PDKP: You are on sitout."], "WHISPER", nil, player)
		end
		db.storedCounter = nextUpdate
	end

	self:CancelTimer(handle_Broadcast, true)
	handle_Broadcast = nil
	handle_MinuteTicker = self:ScheduleTimer(storeTickerCounter, 60, nextUpdate)
	self:RegisterEvent("CHAT_MSG_LOOT")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

function PlantainDKP:StopCounter()
    if notHost() then return end
	LDB.text = "PDKP"

	if not handle_HourTicker and not handle_MinuteTicker and not db.storedCounter then
		print(L["|cff2476ffPDKP:|r Hour counter was not running."])
		return
	end

	self:CancelTimer(handle_HourTicker, true)
	self:CancelTimer(handle_MinuteTicker, true)
	handle_HourTicker = nil
	handle_MinuteTicker = nil
	db.storedCounter = nil
    if db.mode == 1 then
        handle_Broadcast = self:ScheduleRepeatingTimer(broadcastMulti, oneHour)
    end
	print(L["|cff2476ffPDKP:|r Automatic hour counting stopped."])

	self:UnregisterEvent("CHAT_MSG_LOOT")
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

--------------------------------------------------------------------------------
-- Item handling
--

function PlantainDKP:AddItem(player, itemLink)
	if notHost() then return end
	lootDialogOpen = nil
	-- DKP -25%
	local dkp = db.dkp[player] or 0
	local price = 0
	if dkp > 0 then
		price = dkp * 25 / 100
	end
	local newDkp = dkp - price
	db.dkp[player] = newDkp
	local l = db.itemsLooted[player] or 0
	db.itemsLooted[player] = l + 1
	broadcastSingle(player)
	if itemLink then
		print(L["|cff2476ffPDKP:|r %s looted %s, |cffff4488%d|r has been deducted from his DKP."]:format(player, itemLink, price or 0))
        if db.chatloot then
            self:Chat(L["PDKP: %s looted %s, %d has been deducted from his DKP."]:format(player, itemLink, price or 0), "GUILD")
        end
	else
		print(L["|cff2476ffPDKP:|r %s' DKP has been lowered by |cffff4488%d|r, as if he would have looted an item."]:format(player, price or 0))
	end
	popLootQueue()
end

do
	local lootSelf = (LOOT_ITEM_SELF):gsub("%%s", "(.+)")
	local lootSelfMultiple = (LOOT_ITEM_SELF_MULTIPLE):gsub("%%sx%%d", "(.+)x(%%d+)")
	local loot = (LOOT_ITEM):gsub("%%s", "(.+)")
	local lootMultiple = (LOOT_ITEM_MULTIPLE):gsub("%%s receives loot: %%sx%%d", "(.+) receives loot: (.+)x(%%d+)")
	function PlantainDKP:CHAT_MSG_LOOT(_, msg)
		local player, item, quantity = playerName, nil, 1
		if msg:match(lootSelf) then
			item = msg:match(lootSelf)
		elseif msg:match(loot) then
			player, item = msg:match(loot)
		elseif msg:match(lootSelfMultiple) then
			item, quantity = msg:match(lootSelfMultiple)
		elseif msg:match(lootMultiple) then
			player, item, quantity = msg:match(lootMultiple)
		end
		if type(item) == "string" and type(player) == "string" then
			local itemName, itemLink, itemRarity = GetItemInfo(item)
			if not itemRarity or itemRarity < 4 then return end

			-- Find the item ID to make sure it's not a ignored item
			local itemId = select(3, itemLink:find("item:(%d+):"))
			if not itemId then return end
			itemId = tonumber(itemId:trim())
			if type(itemId) ~= "number" or ignoredItems[itemId] then return end

			if type(itemName) == "string" and type(itemLink) == "string" and type(itemRarity) == "number" then
				itemLooted(player, itemLink)
			end
		end
	end
end

--------------------------------------------------------------------------------
-- Hour handling
--

do
	local broadcastTimerHandle = nil
	function PlantainDKP:IncrementHourCounter(input)
		if type(input) == "string" and not input:find("*") then
			input = input:trim():gsub("^%l", string.upper)
			if db.dkp[input] then
				db.dkp[input] = db.dkp[input] + 1
			else
				db.dkp[input] = 1
			end
			db.hours[input] = db.hours[input] and db.hours[input] + 1 or 1
			broadcastSingle(input)
			print(L["|cff2476ffPDKP:|r The hour counter has been |cff44ff44incremented|r by 1 for %s."]:format(input))
		else
			if type(input) == "string" and input ~= "*" then
				print(L["Can't increment the hour counter for %q."]:format(input))
				return
			end
			-- If we're not in a 25man group, we should not process the DKP increase
			-- but we should still restart the hour counter if it's running, so
			-- don't return early here.
			local shouldProcess = true
			if type(input) ~= "string" then
				local n = GetNumRaidMembers()
				if n ~= 25 then
					print("|cffff0000You're not in a 25man raid, please use /bdkp inc * to actually award DKP now.|r")
					shouldProcess = false
				end
			end
			if shouldProcess then
				local dkpTable = {}
				for i = 1, GetNumRaidMembers() do
					local n = GetRaidRosterInfo(i)
					dkpTable[n] = 1 * getDKPMultiplier(n)
				end
				for k, v in pairs(db.sitouts) do
					if not dkpTable[k] then dkpTable[k] = 1 end
				end
				for k, v in pairs(dkpTable) do
					if not db.dkp[k] then
						db.dkp[k] = v
					else
						db.dkp[k] = db.dkp[k] + v
					end
					db.hours[k] = db.hours[k] and db.hours[k] + 1 or 1
				end
				local aAO = db.attendanceAverageOver
				for k in pairs(db.dkp) do
					-- We assume people have full attendance from day 1.
					if not db.attendance[k] then db.attendance[k] = ("1"):rep(aAO) end
					cache_AttendancePercent[k] = nil
					db.attendance[k] = (dkpTable[k] and "1" or "0") .. db.attendance[k]
					db.attendance[k] = db.attendance[k]:sub(1, aAO)
				end
				db.totalRaidHours = db.totalRaidHours + 1
				wipe(dkpTable)
				self:CancelTimer(broadcastTimerHandle, true)
				broadcastTimerHandle = self:ScheduleTimer(broadcastMulti, 5)
				print(L["|cff2476ffPDKP:|r DKP has been |cff44ff44awarded|r for all players in the raid and on the sitout list."])
			end
			if type(input) ~= "string" then
				handle_HourTicker = self:ScheduleTimer("IncrementHourCounter", oneHour)
				self:CancelTimer(handle_MinuteTicker, true)
				handle_MinuteTicker = self:ScheduleTimer(storeTickerCounter, 60, oneHour)
			end
		end
		self:CancelTimer(handle_Broadcast, true)
		if not handle_HourTicker and not handle_MinuteTicker then
			handle_Broadcast = self:ScheduleRepeatingTimer(broadcastMulti, oneHour)
		end
	end
end

function PlantainDKP:DecrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if db.dkp[input] then
			db.dkp[input] = db.dkp[input] - 1
		end
		broadcastSingle(input)
		print(L["|cff2476ffPDKP:|r The hour counter has been |cffff4488decremented|r by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			print(L["Can't decrement the hour counter for %q."]:format(input))
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			dkpTable[n] = true
		end
		for k, v in pairs(db.sitouts) do
			dkpTable[k] = true
		end
		for k, v in pairs(dkpTable) do
			if db.dkp[k] then
				db.dkp[k] = db.dkp[k] - 1
			else
				db.dkp[k] = -1
			end
		end
		wipe(dkpTable)
		broadcastMulti()
		print(L["|cff2476ffPDKP:|r DKP has been |cffff4488decremented|r by 1 for all players in the raid and on the sitout list."])
	end
end

--------------------------------------------------------------------------------
-- Group and guild handling
--

do
	local grouped = nil
	local lastNumRaid = nil
	local lastNumGroup = nil
	function PlantainDKP:RAID_ROSTER_UPDATE()
		local n = GetNumRaidMembers()
		if not lastNumRaid or lastNumRaid ~= n then updateDisplayedNames = true end
		lastNumRaid = n
		if not grouped and n > 0 then
			grouped = true
		elseif grouped and n == 0 then
			self:StopCounter()
			self:ClearSitoutList()
			grouped = nil
		end

		if n > 0 and not handle_HourTicker and db.storedCounter then
			self:StartCounter()
		end
	end
	function PlantainDKP:PARTY_MEMBERS_CHANGED()
		local n = GetNumPartyMembers()
		if not lastNumGroup or lastNumGroup ~= n then updateDisplayedNames = true end
		lastNumGroup = n
		if not grouped and n > 0 then
			grouped = true
		elseif grouped and n == 0 and not UnitInRaid("player") then
			grouped = nil
			self:StopCounter()
			self:ClearSitoutList()
		end
	end
end

do
	-- This is just run on login, since I can't seem to get a reliable hit from
	-- RAID_ROSTER_UPDATE, PLAYER_LOGIN or PLAYER_ENTERING_WORLD on
	-- GetNumRaidMembers ...
	local f = CreateFrame("Frame")
	local total = 0
	f:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed
		if total > 10 then
			if GetNumRaidMembers() > 0 then
				PlantainDKP:RAID_ROSTER_UPDATE()
			end
			self:SetScript("OnUpdate", nil)
			self:Hide()
			f = nil
		end
	end)
end

function PlantainDKP:GUILD_ROSTER_UPDATE()
	wipe(guildRanks)
	wipe(guildMemberList)
	wipe(guildMemberRanks)

	for i = 1, GuildControlGetNumRanks() do
		guildRanks[#guildRanks + 1] = GuildControlGetRankName(i)
	end
	for i = 1, GetNumGuildMembers(true) do
		local name, rank, _, _, _, _, _, _, _, _, class = GetGuildRosterInfo(i)
		if name then
			guildMemberList[name] = class
			guildMemberRanks[name] = rank
		end
	end
	updateDisplayedNames = true
end
function PlantainDKP:GetGuildRanks() return guildRanks end

--------------------------------------------------------------------------------
-- Chat filters
--

--[[do
	local mains = {}
	local addon = CreateFrame("Frame")
	addon:SetScript("OnEvent", function(self, event, ...) self[event](self, ...) end)
	function addon:GUILD_ROSTER_UPDATE()
		for i = 1, GetNumGuildMembers(true) do
			local name, _, _, _, _, _, note = GetGuildRosterInfo(i)
			if name and note then
				mains[note] = name
			end
		end
	end
	function addon:CHAT_MSG_WHISPER(msg, author)
		if mains[author] then
			print("The main of " .. tostring(author) .. " is " .. tostring(mains[author]))
		end
	end
	addon:RegisterEvent("CHAT_MSG_WHISPER")
	addon:RegisterEvent("GUILD_ROSTER_UPDATE")
end]]

local function tablePrinter(t, recipient)
	for i, v in next, t do
		if type(v) == "string" then
			PlantainDKP:Chat(L["PDKP: %s."]:format(v), "WHISPER", nil, recipient)
		elseif type(v) == "table" then
			tablePrinter(v, recipient)
		end
	end
end

local function bidTimeout()
	wipe(tmpBidTable)
	for k in pairs(bidders) do tmpBidTable[#tmpBidTable + 1] = k end
	if #tmpBidTable > 0 then
		table.sort(tmpBidTable, dkpSorter)
		local winner = tmpBidTable[1]
		if #tmpBidTable > 1 and bidders[winner] == bidders[tmpBidTable[2]] then
			local more = true
			while more do
				table.remove(tmpBidTable, #tmpBidTable)
				more = bidders[tmpBidTable[#tmpBidTable]] ~= bidders[winner]
			end
			PlantainDKP:Chat(L["Tie: %s."]:format(table.concat(tmpBidTable, ", ")), "RAID")
		else
			PlantainDKP:Chat(L["%s wins (%d dkp)!"]:format(winner, bidders[winner]), "RAID")
		end
	end
	takingBids = nil
	handle_BidTimer = nil
end

local function newBidder(bidder)
	bidders[bidder] = getModifiedDKP(bidder) or 0
	PlantainDKP:CancelTimer(handle_BidTimer, true)
	handle_BidTimer = PlantainDKP:ScheduleTimer(bidTimeout, 10)
end

function PlantainDKP:CHAT_MSG_WHISPER(event, msg, author)
	if db.mode == 2 then return end
	if msg:find("^bdkp") or msg:find("^pdkp") then
    
        local command = msg:match("^.dkp%s?(.*)")
		local input = command and command:len() > 0 and command:gsub("^%l", string.upper) or author
		local dkp = getDKP(input)
        
		if type(dkp) == "number" then
			local realDkp = getModifiedDKP(input)
			local attendance = getAttendancePercent(input)
			local dkpPercent = getDKPPercent(input)
			local format = L["PDKP: DKP for %s: %s (%s total). 45-hour attendance is at %d%%, which means DKP is worth %d%% of its total value for purposes of bidding."]
			local ret = format:format(input, tostring(realDkp), tostring(dkp), attendance, dkpPercent)
			PlantainDKP:Chat(ret, "WHISPER", nil, author)
		else
			PlantainDKP:Chat(L["PDKP: Unknown command or player not found."], "WHISPER", nil, author)
		end
        
	elseif takingBids and msg:find("^bid") then
		newBidder(author)
	elseif msg == "sitout" then
		PlantainDKP:ProcessSitoutWhisper(author)
	end
end

function PlantainDKP:CHAT_MSG_RAID(event, msg, author)
	if db.mode == 2 then return end
	if author == playerName and msg:find(L["^bid:"]) then
		takingBids = true
		wipe(bidders)
	elseif takingBids and msg:lower() == "bid" then
		newBidder(author)
	end
end
PlantainDKP.CHAT_MSG_RAID_LEADER = PlantainDKP.CHAT_MSG_RAID

function PlantainDKP:Chat(text, chatType, language, destination)
	if ChatThrottleLib then
		ChatThrottleLib:SendChatMessage("BULK", "PlantainDKP", text, chatType, language, destination)
	else
		SendChatMessage(text, chatType, language, destination)
	end
end

--------------------------------------------------------------------------------
-- LDB Plugin Tooltip
--

do
	local names = {}
	local dkpFormat = "%s |cff555555(%d%%)|r"
	local attendanceFormat = "%s |cff555555(%s)|r"
	local attendanceDisplay = "%d%%"

	local realm = GetRealmName()
	local function shouldShow(name)
		if guildMemberList[name] and
			db.disallowRanks[guildMemberRanks[name]] or
			db.disallowClasses[guildMemberList[name]] then
			return nil
		end
		-- Don't show players from other realms.
		local n, r = UnitName(name)
		-- Apparently the second return value can be "" for people from our own realm, not just nil, like wowwiki suggests.
		if n and r and r:trim():len() > 0 and r ~= realm then
			return nil
		end
		return true
	end

	function LDB.OnTooltipShow(tt)
		tt:AddLine("PlantainDKP")
		if db.mode == 2 and db.lastDKPUpdate then
			tt:AddDoubleLine(L["Updated"], date("%d. %b@%H:%M", db.lastDKPUpdate), 1,1,0, 1,1,1)
		end

		if IsAltKeyDown() or updateDisplayedNames then
			wipe(names)
			if not IsAltKeyDown() and GetNumRaidMembers() > 0 then
				for i = 1, GetNumRaidMembers() do
					local n = GetRaidRosterInfo(i)
					if n and shouldShow(n) then
						names[#names + 1] = n
					end
				end
			elseif not IsAltKeyDown() and GetNumPartyMembers() > 0 then
				table.insert(names, playerName)
				for i = 1, GetNumPartyMembers() do
					local n = UnitName("party" .. i)
					if n and shouldShow(n) then
						names[#names + 1] = n
					end
				end
			else
				for k in pairs(guildMemberList) do
					if shouldShow(k) then
						local d = getDKP(k)
						if d then names[#names + 1] = k end
					end
				end
			end
			updateDisplayedNames = nil
			if db.tooltipSortMethod == "dkp" then
				table.sort(names, dkpSorter)
			else
				table.sort(names, attendanceSorter)
			end
		end

		if #names > 0 then
			tt:AddLine(" ")
			if db.tooltipSortMethod == "dkp" then
				tt:AddDoubleLine(L["Player"], L["DKP"], 1,1,1, 1,1,1)
			else
				tt:AddDoubleLine(L["Player"], L["Attendance"], 1,1,1, 1,1,1)
			end
			for i, v in next, names do
				local totalDkp = getDKP(v)
				if totalDkp and totalDkp > 0 then
					local realDkp = getModifiedDKP(v)
					local attendance = getAttendancePercent(v)
					if v == playerName then
						if db.tooltipSortMethod == "dkp" then
							tt:AddDoubleLine(dkpFormat:format(v, attendance), realDkp, 1,0,0, 1,0,0)
						else
							tt:AddDoubleLine(attendanceFormat:format(v, realDkp), attendanceDisplay:format(attendance), 1,0,0, 1,0,0)
						end
					else
						if db.tooltipSortMethod == "dkp" then
							tt:AddDoubleLine(dkpFormat:format(coloredName[v], attendance), realDkp, 1,1,1, 1,1,1)
						else
							tt:AddDoubleLine(attendanceFormat:format(coloredName[v], realDkp), attendanceDisplay:format(attendance), 1,1,1, 1,1,1)
						end
					end
				end
			end
		end

		if next(db.sitouts) then
			tt:AddLine(" ")
			local tmp = {}
			for k in pairs(db.sitouts) do tmp[#tmp + 1] = coloredName[k] end
			tt:AddLine(L["On sitout"], 1, 1, 1)
			tt:AddLine(table.concat(tmp, ", "), 1, 1, 1, 1)
		end
        
		if db.mode == 2 and db.host then
            tt:AddLine(" ")
			tt:AddDoubleLine(L["Role: Client"], L["Server: %s"]:format(coloredName[db.host]), 1,1,1, 1,1,1)
		end
	end
end

--------------------------------------------------------------------------------
-- Sitout list handling
--

function PlantainDKP:ProcessSitoutWhisper(player)
	if not guildMemberList[player] then
		print(L["|cff2476ffPDKP:|r %q tried to add himself to our sitout list, but he's |cffff4488not in our guild|r."]:format(player))
		return
	end
	if db.sitouts[player] then
		self:RemoveSitoutPlayer(player)
	else
		self:AddSitoutPlayer(player)
	end
end

function PlantainDKP:ClearSitoutList()
	wipe(db.sitouts)
end

function PlantainDKP:AddSitoutPlayer(name)
	db.sitouts[name] = true
	print(L["|cff2476ffPDKP:|r %q |cff44ff44added|r to the sitout list."]:format(name))
	self:Chat(L["PDKP: You have been ADDED to the sitout list. Whisper me again to remove yourself."], "WHISPER", nil, name)
end

function PlantainDKP:RemoveSitoutPlayer(name)
	db.sitouts[name] = nil
	print(L["|cff2476ffPDKP:|r %q |cffff4488removed|r from the sitout list."]:format(name))
	self:Chat(L["PDKP: You have been REMOVED from the sitout list. Whisper me again to add yourself."], "WHISPER", nil, name)
end

